from __future__ import annotations

from datetime import timedelta

from django.conf import settings as django_settings
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.utils.timezone import now as tz_now
from django.utils.translation import gettext_lazy as _lazy

from .trigger_subscription import TriggerSubscription


class TriggerEvent(models.Model):

    user = models.ForeignKey(
        django_settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="zapier_trigger_events",
    )
    trigger = models.CharField(
        max_length=100, help_text=_lazy("The name of the Zapier trigger.")
    )
    subscription = models.ForeignKey(
        TriggerSubscription,
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        related_name="trigger_events",
        help_text=_lazy(
            "The subscription to which the event was posted (null for polling events)."
        ),
    )
    http_method = models.CharField(
        max_length=4,
        help_text=_lazy("How the data was sent to Zapier - via GET, or POST."),
    )
    event_data = models.JSONField(
        blank=True,
        null=True,
        encoder=DjangoJSONEncoder,
        help_text=_lazy("JSON data sent to Zapier."),
    )
    started_at = models.DateTimeField(default=tz_now, blank=True, null=True)
    finished_at = models.DateTimeField(blank=True, null=True)
    status_code = models.IntegerField(
        help_text=_lazy("HTTP status code received from Zapier")
    )

    def __str__(self) -> str:
        return f"'{self.trigger}' event #{self.id}"

    @property
    def duration(self) -> timedelta | None:
        if not self.finished_at:
            return None
        return self.finished_at - self.started_at



from typing import Callable

from rest_framework.permissions import BasePermission
from rest_framework.request import Request

from .settings import STRICT_MODE


class IsZapier(BasePermission):
    """
    Enforce STRICT_MODE.

    The User-Agent header for all requests from Zapier is set to
    "Zapier". This is trivial to spoof, so this check doesn't increase
    security in any way, but it's a useful signal in case of casual
    abuse. (Casual = someone probing with no real intent.)

    """

    def has_permission(self, request: Request, view: Callable) -> bool:
        if not STRICT_MODE:
            return True
        return request.headers.get("user-agent", "") == "Zapier"



from __future__ import annotations

from django.urls import path

from .views import TriggerView, auth_check

app_name = "zapier_triggers"


urlpatterns = [
    # GET - the auth check URL
    path("auth/", auth_check, name="auth_check"),
    # GET - the "list" trigger endpoint
    path("<str:trigger>/", TriggerView.as_view(), name="list"),
    # POST - the "subscribe" endopint
    path("<str:trigger>/subscriptions/", TriggerView.as_view(), name="subscribe"),
    # DELETE - "unsubscribe" endpoint
    path(
        "<str:trigger>/subscriptions/<uuid:subscription_id>/",
        TriggerView.as_view(),
        name="unsubscribe",
    ),
]



import importlib
from typing import Callable, Type

from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured
from rest_framework.authentication import BaseAuthentication

from .types import TriggerViewFunc

# read in settings that have been overridden in django settings.py
_settings = getattr(django_settings, "ZAPIER_TRIGGERS", {})
_settings.setdefault("AUTHENTICATOR", None)
_settings.setdefault("STRICT_MODE", not django_settings.DEBUG)
_settings.setdefault("HOOK_URL_KEY", "hookUrl")
_settings.setdefault("TRIGGERS", {})

# set to True to reject requests that don't come from Zapier
STRICT_MODE = _settings["STRICT_MODE"]

# the JSON key used to extract the REST Hook url from inbound post ("hookUrl")
HOOK_URL_KEY = _settings["HOOK_URL_KEY"]

# map of trigger_name: get_data_func - all triggers must be configured
TRIGGERS = _settings["TRIGGERS"]


def import_from_path(path: str) -> Type | Callable:
    """Import function from string path."""
    module, func_or_klass = path.rsplit(".", 1)
    m = importlib.import_module(module)
    return getattr(m, func_or_klass)


def get_authenticator() -> BaseAuthentication:
    """Return authentication class imported from AUTHENTICATOR setting."""
    if not (path := _settings["AUTHENTICATOR"]):
        raise ImproperlyConfigured("Missing AUTHENTICATOR.")
    return import_from_path(path)


def get_trigger(trigger: str) -> TriggerViewFunc:
    """Return view data function configured for the trigger."""
    try:
        return import_from_path(_settings["TRIGGERS"][trigger])
    except KeyError:
        raise ImproperlyConfigured("Missing trigger view function.")


def trigger_exists(trigger: str) -> bool:
    """Check that the trigger is configured."""
    return trigger in _settings["TRIGGERS"]



from typing import Any

from django.core.serializers.json import DjangoJSONEncoder
from django.http import JsonResponse as DjangoJsonResponse


class JsonResponse(DjangoJsonResponse):
    def __init__(self, data: list | dict, **kwargs: Any) -> None:
        kwargs.setdefault("encoder", DjangoJSONEncoder)
        super().__init__(data, **kwargs)



from __future__ import annotations

import json
import logging
from functools import wraps
from typing import Any
from uuid import UUID

from django.http import HttpResponseNotFound
from django.shortcuts import get_object_or_404
from django.utils.timezone import now as tz_now
from rest_framework.decorators import (
    api_view,
    authentication_classes,
    permission_classes,
)
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

from .models import TriggerEvent, TriggerSubscription
from .permissions import IsZapier
from .response import JsonResponse
from .settings import HOOK_URL_KEY, get_authenticator, get_trigger, trigger_exists
from .subscription import subscribe, unsubscribe
from .types import TriggerData, TriggerViewMethod

logger = logging.getLogger(__name__)

AUTHENTICATOR = get_authenticator()


@api_view(["GET"])
@authentication_classes([AUTHENTICATOR])
@permission_classes([IsAuthenticated, IsZapier])
def auth_check(request: Request) -> Response:
    """Support Zapier auth check."""
    logger.debug("Successful authentication request.")
    return JsonResponse({"connectionLabel": request.user.username}, status=200)


def trigger_method(view_method: TriggerViewMethod) -> TriggerViewMethod:
    """Return 404 if trigger passed to view method does not exist."""

    @wraps(view_method)
    def decorated(
        view: TriggerView, request: Request, trigger: str, *args: Any, **kwargs: Any
    ) -> JsonResponse | HttpResponseNotFound:
        if trigger_exists(trigger):
            return view_method(view, request, trigger, *args, **kwargs)
        return HttpResponseNotFound("Unknown trigger request.")

    return decorated


class TriggerView(APIView):
    """
    Manage Zapier REST Hook subscriptions and route polling requests.

    This is a DRF APIView that maps the POST/DELETE/GET methods to the
    Zapier REST hook trigger functions subscribe, unsubscribe, list.

    The AUTHENTICATION_CLASS is read in from the settings.py, and allows
    client applications to control which authentication is used. This
    supports all the authentication mechanisms available to DRF (Token,
    Basic, Session) and its third party extensions (OAuth2 etc.), which
    map onto the auth models that Zapier itself supports. NB Zapier only
    supports one authentication mechanism per app, so this is a single
    value, not a list.

    The POST/DELETE method create / delete REST Hook subscriptions. The
    GET method calls the function configured in IST_FUNCS which can
    either return real data (for a polling trigger), or static sample
    data for a REST Hook.

    The IsZapier permissions class checks for the "Zapier" user-agent if
    STRICT_MODE is enabled.

    The @check_trigger decorator returns a 404 response if the trigger
    parameter is not configured as a LIST_FUNC key.

    """

    authentication_classes = [AUTHENTICATOR]
    permission_classes = [IsAuthenticated, IsZapier]

    def get_trigger_data(self, request: Request, trigger: str) -> TriggerData:
        """Call the configured trigger view function."""
        return get_trigger(trigger)(request)

    @trigger_method
    def get(self, request: Request, trigger: str) -> JsonResponse:
        """
        Fetch trigger data.

        Every trigger must have a func configured that will return data
        as the "list" function. For polling triggers this should be the
        real polling trigger data; for resthook triggers this can be
        some static sample data - it is only used for the Zap UI.

        """
        logger.debug("Fetching data for '%s' trigger.", trigger)
        started_at = tz_now()
        event_data = self.get_trigger_data(request, trigger)
        # we only record if data exists.
        if event_data:
            TriggerEvent.objects.create(
                user=request.user,
                trigger=trigger,
                event_data=event_data,
                http_method="GET",
                started_at=started_at,
                finished_at=tz_now(),
                status_code=200,
            )
        return JsonResponse(data=event_data, status=200, safe=False)

    @trigger_method
    def post(self, request: Request, trigger: str) -> JsonResponse:
        """Create a new webhook subscription."""
        hook_url = json.loads(request.body.decode()).get(HOOK_URL_KEY)
        subscription = subscribe(request.user, trigger, target_url=hook_url)
        # response JSON is stored in `bundle.subscribeData`
        return JsonResponse({"id": str(subscription.uuid)}, status=201)

    @trigger_method
    def delete(
        self, request: Request, trigger: str, subscription_id: UUID
    ) -> JsonResponse:
        """Deactivate an existing webhook subscription."""
        subscription = get_object_or_404(TriggerSubscription, uuid=subscription_id)
        unsubscribe(subscription)
        return JsonResponse({}, status=204)



# type alias for the "list" view functions
from typing import Callable, TypeAlias

from django.http import HttpResponseNotFound
from rest_framework.request import Request

from zapier.triggers.response import JsonResponse

TriggerData: TypeAlias = list[dict]
TriggerViewFunc: TypeAlias = Callable[[Request], TriggerData]
TriggerViewMethod: TypeAlias = Callable[..., JsonResponse | HttpResponseNotFound]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zapier',
 'zapier.triggers',
 'zapier.triggers.migrations',
 'zapier.triggers.models']

package_data = \
{'': ['*']}

install_requires = \
['django>=3.2,<5.0',
 'djangorestframework',
 'python-dateutil',
 'pytz',
 'requests']

setup_kwargs = {
    'name': 'django-zapier-trigger',
    'version': '1.0b1',
    'description': 'Django (DRF) backed app for managing Zapier triggers.',
    'long_description': '# Django Zapier Triggers\n\nDjango (DRF) app for managing Zapier triggers.\n\n### Version support\n\nThis app supports Django 3.2+ and Python 3.10+.\n\n## Background\n\nThis app provides the minimal scaffolding required to support Zapier triggers in your Django application. It is based on DRF.\n\nIn addition to the `zapier.triggers` Django app, this project includes two additional applications: a complete Zapier CLI application that you can publish to Zapier, and a Demo project that provides the Django support for it. With these two projects you have a complete end-to-end Zapier integration.\n\n## Zapier Triggers\n\nA Zapier trigger is an event source for Zapier workflows ("Zaps"), that can operate in one of two modes - "Instant", or "Polling". Either way the net result is that JSON data objects are received by Zapier and can be used as the first step in a Zap.\n\nThere is a _lot_ of documentation online from Zapier about how to create a trigger, and I would strongly recommend reading it before attempting to build your own. Here are a couple of good articles to start with:\n\n- https://platform.zapier.com/docs/start\n- https://platform.zapier.com/cli_tutorials/getting-started\n\n### Prequisites\n\nIf you want to run the end-to-end demo you will need:\n\n1. A Zapier account\n2. The Zapier CLI\n3. ngrok, or some equivalent tunnelling software\n\n## What\'s in the box?\n\nThe core implementation detail of this package is the `TriggerView`. This is a DRF `APIView` class that handles `GET`, `POST`, and `DELETE` methods, mapping them to the Zapier trigger methods for polling ("list"), susbscribe and unsubscribe functions.\n\n### `GET /triggers/{{trigger}}/`\n\nWhen Zapier makes a `GET` request to your application endpoint one of two things is happening. For a REST Hook ("Instant") trigger this is request sample data that Zapier can use to create its Zap builder UI. If your trigger is a push ("Instant") then you can just return static data - as long as it conforms to the same schema as real data. The `demo.views.new_book` view demonstrates this.\n\nIf your trigger is a polling trigger then this endpoint should return real data - the `demo.views.new_film` view is an example of this.\n\nThe view returns a `200` status code.\n\n### `POST /triggers/{{trigger}}/subscriptions/`\n\nWhen Zapier makes a `POST` request it is expecting to create a new webhook (rebranded "REST Hook" by Zapier) susbscription. This is handled automatically by the view, which creates a new `TriggerSubscription` object for the user + trigger combination, and returns the `uuid` property to Zapier, which stores it in its `bundle.subscriptionData.id` property.\n\nThe view returns a `201` status code.\n\n### `DELETE /triggers/{{trigger}}/subscriptions/{{subscription_id}}`\n\nWhen Zapier makes a `DELETE` request it is expecting to delete the subscription identified by the `subscription_id` value, which maps to the `uuid` property. We do not delete the subscription but instead mark it as "inactive". This is because we record all of the event data that is sent by a trigger subscription, and we we want to keep this for a period for auditing purposes. If a new `POST` request is made for the same user + trigger combination the subscription is reactivated.\n\nThe view returns a `204` status code.\n\n## Settings\n\nThe settings are all read in from the Django setting `ZAPIER_TRIGGER`, which is a dict containing the following keys:\n\n* `STRICT_MODE`\n\nIf True this will check that incoming requests come from Zapier. Defaults to `not DEBUG` - so it\'s on in production, and off in local development.\n\n* `HOOK_URL_KEY`\n\nThe JSON key used to extract the Zapier subscription URL endpoint in the body of the `POST` request - defaults to `hookUrl`.\n\n* `TRIGGERS`\n\nThis is a dict containing the name of the trigger and a string path to a view-like function that must accept a single `Request` arg and return a list of JSON-serializable dict objects. Every trigger that your Zapier app supports must be in this setting - otherwise any request made to `/triggers/{{trigger}}` will return a `404`.\n\n## Demo + zapier-app\n\nThe easiest way to work out how this all fits together is to run the demo app and push the zapier-app to Zapier under your own account.\n',
    'author': 'YunoJuno',
    'author_email': 'code@yunojuno.com',
    'maintainer': 'YunoJuno',
    'maintainer_email': 'code@yunojuno.com',
    'url': 'https://github.com/yunojuno/django-zapier-trigger',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import logging
from uuid import uuid4

from django.conf import settings as django_settings
from django.db import models
from django.utils.timezone import now as tz_now
from django.utils.translation import gettext_lazy as _lazy

logger = logging.getLogger(__name__)


class TriggerSubscriptionQuerySet(models.QuerySet):
    def active(self) -> TriggerSubscriptionQuerySet:
        """Filter active subscriptions."""
        return self.filter(subscribed_at__isnull=False, unsubscribed_at__isnull=True)

    def trigger(self, trigger: str) -> TriggerSubscriptionQuerySet:
        """Filter by trigger."""
        return self.filter(trigger=trigger)


class TriggerSubscription(models.Model):
    """
    Implementation of REST Hooks to store webhook subscriptions.

    See http://resthooks.org/docs/

    """

    uuid = models.UUIDField(
        default=uuid4,
        help_text=_lazy("Public ID"),
    )
    user = models.ForeignKey(
        django_settings.AUTH_USER_MODEL,
        related_name="zapier_subscriptions",
        on_delete=models.CASCADE,
    )
    trigger = models.CharField(
        max_length=50,
        db_index=True,
        help_text=_lazy("The name of the trigger event to subscribe to."),
    )
    target_url = models.URLField(
        help_text=_lazy("The webhook URL to which any payload will be POSTed.")
    )
    subscribed_at = models.DateTimeField(
        default=tz_now,
        help_text=_lazy("Timestamp marking when the initial subscribe event occurred."),
    )
    unsubscribed_at = models.DateTimeField(
        default=None,
        null=True,
        blank=True,
        help_text=_lazy("Timestamp marking when the unsubscribe event occurred."),
    )

    objects = TriggerSubscriptionQuerySet.as_manager()

    def __str__(self) -> str:
        return f"Subscription #{self.id} ('{self.trigger}')"

    @property
    def is_active(self) -> bool:
        """Return True if the subscription is active."""
        return self.subscribed_at and not self.unsubscribed_at

    @property
    def is_inactive(self) -> bool:
        """Return False if the subscription is active."""
        return self.subscribed_at and self.unsubscribed_at

    def serialize(self) -> dict:
        """Serialize object as a JSON-serializable dict."""
        return {
            "uuid": self.uuid,
            "user_id": self.user.pk,
            "trigger": self.trigger,
            "url": self.target_url,
            "active": self.is_active,
        }

    def unsubscribe(self) -> None:
        self.target_url = ""
        self.unsubscribed_at = tz_now()
        self.save(update_fields=["target_url", "unsubscribed_at"])

    def resubscribe(self, target_url: str) -> None:
        self.target_url = target_url
        self.subscribed_at = tz_now()
        self.unsubscribed_at = None
        self.save(update_fields=["target_url", "subscribed_at", "unsubscribed_at"])
